define(['jquery', 'underscore', 'backbone', 'uicontrols'], ($, _, Backbone, $c) => {
  Backbone.Collection.prototype.totalDestroy = function () {
    let 	i = this.models.length;

    while (i-- > 0) {
      this.models[i].clear({ silent: true });
      this.models[i].trigger('clear');
    }
    this.reset();
  };

  var App = {
    model: {},
    view: {},
    collection: {},
    helpers: {
      // @param {String} chain - property name
      // @param {Object} root - property object
      // @return {Object} result property if founded
      property(chain, root) {
        const 	names = chain.split('.');
        let i = -1; const len = names.length; let
          ref = root || window;

        while (i++, i < len) {
          ref = ref[names[i]];
          if (ref == undefined) {
            break;
          }
        }
        return ref;
      },
      // @memberOf App.helpers
      // @param {Function}
      // @return {Var | null}
      tryBack(cb) {
        try {
          return cb();
        } catch (e) {
          return null;
        }
      },
      withZero(n) {
        return (n < 10 ? '0' : '') + n;
      },
      // @memberOf App.helpers
      // @param {String} template
      // @return {Function}
      timeFormatter(template) {
        const 	parts = template.split(/(%\w)/g);
        _callbacks = {
          '%h': function (d) { return App.helpers.withZero(d.getHours()); },
          '%m': function (d) { return App.helpers.withZero(d.getMinutes()); },
          '%s': function (d) { return App.helpers.withZero(d.getSeconds()); },
          '%D': function (d) { return App.helpers.withZero(d.getDate()); },
          '%M': function (d) { return App.helpers.withZero(d.getMonth() + 1); },
          '%Y': function (d) { return d.getFullYear(); },
        };

        return function (date) {
          var 	date = (typeof (date) === 'number' ? new Date(date * 1000) : date) || new Date();
          let str = '';

          for (let i = 0; i < parts.length; i++) {
            if (_callbacks[parts[i]]) {
              str += _callbacks[parts[i]](date);
            } else {
              str += parts[i];
            }
          }
          return str;
        };
      },
      // @memberOf App.helpers - get total time between Dates objects
      timeLeft(date1, date2) {
        let 	sec = ~~((date2 - date1) / 1000);
        let min = ~~(sec / 60);
        let hour = ~~(min / 60);
        const days = ~~(hour / 24);

        hour -= 24 * days;
        min -= 60 * (hour + 24 * days);
        sec -= 60 * (min + 60 * (hour + 24 * days));

        return {
          d: days,
          h: hour,
          m: min,
        };
      },
      // DEPRICATED (correct for russian only)
      // @memberOf App.helpers - plural function
      plural(a, form1, form2, form3) {
        if (a % 10 == 1 && a % 100 != 11) return form1;
        if (a % 10 >= 2 && a % 10 <= 4 && (a % 100 < 10 || a % 100 >= 20)) return form2;
        return form3;
      },
      // plurals: function(a, forms){
      // 	if(a % 10 == 1 && a % 100 != 11){
      // 		return forms[0];
      // 	}else if(a % 10 >= 2 && a % 10 <= 4 && ( a % 100 < 10 || a % 100 >= 20)){
      // 		return forms[1];
      // 	}else{
      // 		return forms[2] || forms[1];
      // 	}
      // },
      plurals(n, lang, forms) {
        let i = 0;

        if (lang == 'ru' || lang == 'uk') {
          i = (n % 10 == 1 && n % 100 != 11 ? 0 : n % 10 >= 2 && n % 10 <= 4 && (n % 100 < 10 || n % 100 >= 20) ? 1 : 2);
        } else if (lang === 'jp') { // jp
          i = 0;
        } else { // en, ge, fr
          i = n != 1 ? 1 : 0;
        }
        return forms[i];
      },
      formatTime(ms, _units) {
        const 	units = _units || ['ms', 's', 'm', 'h', 'd'];
        const mult = [1000, 60, 60, 24];
        let tail;
        let head = ms;
        let out = '';
        let i = 0;

        for (; head >= mult[i] && i < 4; i++) {
          tail = head % mult[i];
          head = (head - tail) / mult[i];
          if (tail) out = tail + units[i] + out;
        }
        return head + units[i] + out;
      },
      formatBytes(byts, units) {
        var 	units = units || ['b', 'kb', 'mb', 'gb', 'tb'];
        let bytes = byts;
        let i;

        for (i = 0; bytes >= 1024 && i < 4; i++) {
          bytes /= 1024;
        }
        // bytes must be cell Number.isInteger(bytes)
        return (i > 0 ? bytes.toFixed(2) : bytes) + units[i];
      },
      i18transformer(pattern, conf) {
        return pattern.replace(/\{([^\}]+)\}/g, (fragment, match, pos) => {
          if (conf[match]) {
            return conf[match];
          }
          const args = match.split('|');

          if (args.length == 4 && conf[args[0]]) {
            return App.helpers.plural(conf[args[0]], args[1], args[2], args[3]);
          }
          return '_';
        });
      },
      pd(e) {
        e.preventDefault();
      },
      sp(e) {
        e.stopPropagation();
      },
      _(s) {
        return App.i18n._(s) || s;
      },
      notify(title, mes) {
        new $c.BasePopupView({
          title: `<strong>${_.escape(title)}</strong>`,
          class: 'dec_wt-popup',
          content:
                    '<form class="dwc-form" data-co="form">'
                    + `<div class="dwc-form_offset1">${_.escape(mes)}</div>`
                    + '<div class="dwc-form_btn-wrap">'
                    + `<button class="dwc_btn" data-co="submit">${App.helpers._('Close')}</button>`
                    + '</div>'
                    + '</form>'
                    + '',
          onopen(view) {
            view.controls.form.on('submit', (e) => {
              e.stopPropagation();
              e.preventDefault();
              view.close();
            });
          },
          onclose(view) {},
        }).open();
      },
      confirm(title, mes, onSubmit, onCancel) {
        new $c.BasePopupView({
          title: `<strong>${_.escape(title)}</strong>`,
          class: 'dec_wt-popup',
          content:
                    '<form class="dwc-form" data-co="form">'
                    + `<div class="dwc-form_offset1">${_.escape(mes)}</div>`
                    + '<div class="dwc-form_btn-wrap">'
                    + `<button class="dwc_btn" data-co="submit">${App.helpers._('Yes')}</button>`
                    + `<button class="dwc_btn" data-co="cancel">${App.helpers._('No')}</button>`
                    + '</div>'
                    + '</form>'
                    + '',
          moveable: true,
          onopen(view) {
            view.controls.form.on('submit', (e) => {
              e.stopPropagation();
              e.preventDefault();
              onSubmit && onSubmit();
              view.close();
            });
            view.controls.cancel.on('click', (e) => {
              onCancel && onCancel();
              view.close();
            });
          },
          onclose(view) {},
        }).open();
      },
      localizeAttr(el) {
        const 	i18AttrNodes = el.querySelectorAll('[data-i18-attr]');
        let key;

        // Localize attribute values (title:scannedHelp)
        for (let i = 0; i < i18AttrNodes.length; i++) {
          key = i18AttrNodes[i].dataset.i18Attr.split(':');
          key[0] && i18AttrNodes[i].setAttribute(key[0], this._(key[1]));
        }
      },
      localize(el) {
        const 	i18Nodes = el.querySelectorAll('[data-i18]');
        let key;

        for (let i = 0; i < i18Nodes.length; i++) {
          key = i18Nodes[i].dataset.i18;

          i18Nodes[i].textContent = _.unescape(App.i18n._(key));
        }
      },
      generateBulkPage() {
        return 'data:text/html;charset=utf-8,'
                    + `<html style="background:#fff;font:13px/15px Arial;color:#111;"><head></head><body style="height:100%;margin:0;text-align:center;"><div style="display:inline-block;vertical-align:middle;"><div style="font-weight:bold;margin:0 0 20px 0;">${App.helpers._('server_was_crashed')}</div><div>${
                      App.helpers._('turn_on_httpd').replace(/\[b\]([^\[]+)\[\/b\]/, '<span style="font-family:\'Courier New\';background-color:#ebebeb;padding:0 5px;">$1</span>')
                    // '<html style="background:#fff;font:13px/15px Arial;color:#111;"><head></head><body style="height:100%;margin:0;text-align:center;"><div style="display:inline-block;vertical-align:middle;"><div style="font-weight:bold;">' + App.helpers._('server_was_crashed') +
                    }</div></div><div style="display:inline-block;height:100%;width:0;vertical-align:middle;"></div></body></html>`;
      },
      d($node, key, value) {
        if (value != undefined) {
          $node.setAttribute(`data-${key}`, value);
          if ($node.dataset) $node.dataset[key] = value; // IE10 not support dataset collection
        } else {
          if ($node.dataset) return $node.dataset[key];
          return $node.getAttribute(`data-${key}`);
        }
      },
      patterns: {
        // At dev, for localization perposing
        parseStrWithNodes(template, replaceValues) {
          const frag = document.createDocumentFragment();
          let pos = 0;
          const pattern = /\[([^\]]+)\]([^\[]+)\[\/\w+\]/g;
          let node; let list; let i; let buf; let
            res;

          while (res = pattern.exec(template)) {
            if (res.length == 3) {
              list = res[1].split('|');
              node = document.createElement(list.shift());
              i = list.length;
              while (i-- > 0) {
                buf = list[i].indexOf('=');
                var key; var
                  value;
                if (buf != -1) {
                  key = list[i].substr(0, buf);
                  value = list[i].substr(buf + 1);
                } else {
                  key = list[i];
                  value = '';
                }
                if (replaceValues && (value in replaceValues)) {
                  value = replaceValues[value];
                }
                node.setAttribute(key, value);
              }

              node.textContent = res[2];
            }

            frag.appendChild(document.createTextNode(template.substring(pos, res.index)));
            frag.appendChild(node);
            pos = res.index + res[0].length;
          }

          frag.appendChild(document.createTextNode(template.substr(pos)));
          return frag;
        },
      },
    },
    xhr: {
      _isJsonMIME: /application\/json/,
      safeJSON(text) {
        try {
          return JSON.parse(text);
        } catch (e) {
          return null;
        }
      },
      // @member of $xhr - send ajax request
      // @param {Function} oncomplete - callback
      // @param {Bool} forceJSON - force convert JSON
      _create(oncomplete, forceJSON) {
        const xhr = new XMLHttpRequest();

        xhr.onerror = function (e) {
          App.bus.trigger('disconnect');
        };
        xhr.onreadystatechange = function () {
          if (xhr.readyState == 4) {
            const 	contentType = xhr.getResponseHeader('Content-Type');
            const rez = {
              status: xhr.status,
              sessionComplete: false,
              ec: undefined, // aka Error Code
            };
            let response;

            // if(xhr.status == 0){} // also fail

            if (this._isJsonMIME.test(contentType) && xhr.responseText || forceJSON) {
              response = this.safeJSON(xhr.responseText);
              rez.ec = response && response.error && response.error.code;

              if (rez.ec == 'EC_NOT_AUTHORIZED' || rez.ec == 'EC_AUTH_FAILED' || rez.ec == 'EC_INVALID_TOKEN') {
                rez.sessionComplete = true;
                App.user && App.user.set('user', ''); // also you can do window.location.reload();
              }
            } else {
              response = xhr.responseText;
            }

            oncomplete && oncomplete(response, rez);
          }
        }.bind(this);
        xhr.overrideMimeType && xhr.overrideMimeType('application/json');
        return xhr;
      },
      _serialized(params) {
        return params && JSON.stringify(params);
      },
      _urlEncode(params) {
        const 	urlencoded = [];

        for (const key in params) {
          params.hasOwnProperty(key) && urlencoded.push(`${key}=${encodeURIComponent(params[key])}`);
        }
        return urlencoded.join('&');
      },
      post(url, params, oncomplete, resFormat) {
        const xhr = this._create(oncomplete, resFormat == 'json');
        xhr.open('POST', url, true);
        xhr.setRequestHeader('content-type', 'application/json');
        xhr.send(this._serialized(params));
        return xhr;
      },
      get(url, params, oncomplete, resFormat) {
        const 	xhr = this._create(oncomplete, resFormat == 'json');
        const urlparams = this._urlEncode(params);

        if (urlparams) {
          url += `?${urlparams}`;
        }

        xhr.open('GET', url, true);
        xhr.setRequestHeader('content-type', 'application/json');
        xhr.send(null);
        return xhr;
      },
      objPost(conf) {
        const xhr = this._create(conf.oncomplete, conf.resFormat == 'json');
        xhr.open('POST', conf.url, true);
        xhr.setRequestHeader('content-type', conf.contentType || 'application/json');
        // xhr.responseType = 'arraybuffer'; // ?? test
        xhr.send(conf.params || this._serialized(params));
        return xhr;
      },
    },
  };

  App.router = new (Backbone.Router.extend({
    routes: { // map url routes to contained methods
      main: 'main',
      login: 'login',
      threats: 'threats',
      settings: 'settings',
      'settings/:id': 'settings',
      help: 'help',
      info: 'info',
      '*path': 'login', // default Route
    },
    // Attetion! would be fired after page load
    main() {},
    login() {},
    settings() {},
  }))();

  // Event "bus"
  App.bus = _.extend({}, Backbone.Events);

  App.model.lic = Backbone.Model.extend({
    defaults: {
      es_mode: '',
    },
  });

  App.model.auth = Backbone.Model.extend({
    defaults: {
      user: '',
      password: '',
    },
    url: '/api/10.2/login',
    parse(response, options) {
      return response;
    },
    /*    autenticate: function(login, pass){
                this.save({
                    user: login,
                    password: pass
                }, {
                    type: 'POST',
                    contentType: 'application/json',
                    //dataType: "text", // if serever not return JSON object success handler never fired!
                    success: function(m, r, obj){
                        this.set('password', '***', {silent: true});
                        // var code = App.helpers.property('error.code', r);
                        // code && m.trigger('fail', code);
                    }.bind(this),
                    error: function(m, r, obj){
                        var 	code = App.helpers.property('responseJSON.error.code', r);

                        m.trigger('fail', code);
                        r.status == 0 && App.bus.trigger('disconnect');
                        this.set('password', '***', {silent: true});
                    }.bind(this),
                    wait: true
                });
            }, */
    autenticate(login, pass) {
      this.save({
        user: login,
        password: null,
      }, {
        type: 'POST',
        contentType: 'application/json',
        data: JSON.stringify({
          user: login,
          password: pass,
        }),
        success(m, r, obj) {
          // This is model
          m.set('password', '***', { silent: true });
        },
        error(m, r, obj) {
          const 	code = App.helpers.property('responseJSON.error.code', r);

          m.trigger('fail', code);
          r.status == 0 && App.bus.trigger('disconnect');
          m.set('password', '***', { silent: true });
        },
        wait: true,
        // dataType: "text", // if serever not return JSON object success handler never fired!
        // success: function(m, r, obj){
        // 	var code = App.helpers.property('error.code', r);
        // 	code && m.trigger('fail', code);
        // },
      });
    },
    logout() {
      this.save({
        user: '',
        password: '',
      }, {
        data: '',
        url: '/api/10.2/logout',
        type: 'GET',
        contentType: 'application/json',
        // dataType: "text", // if server not return JSON object success handler never fired!
        success(m, r, obj) {},
        error(m, r, obj) {
          r.status == 0 && App.bus.trigger('disconnect');
        },
        wait: true,
      });
    },
    // @memberOf App.model.auth - check page autorization // App.model.auth.checkAutorization()
    // @param {Function} cb
    checkAutorization(cb) {
      // Token passport
      const $xhr = App.xhr.get('/api/10.2/whoami', {}, (d, r) => {
        const 	user = App.helpers.property('whoami.user', d) || '';
        const code = App.helpers.property('error.code', d);

        this.setf('user', user);
        cb && cb(user);
      }, 'json');

      setTimeout(() => {
        $xhr.abort();
      }, 10000); // waiting timeout
    },
    // forse set
    setf(key, value) {
      const options = { silent: true };
      this.set(key, value, options);
      this.trigger(`change:${key}`, this, value, options);
      this.trigger('change', this, options);
    },
  });

  // Internationalizate model // TODO maybe rename to indexModel, because it use at whole app.
  App.model.I18n = Backbone.Model.extend({
    defaults: {
      lang: 'en',
      maxCheckFileSize: 24,
    },
    // @param {String} key - vocabular key
    // @return {String} - translated value or key
    _(key) {
      // DEPRICATED
      //		var 	voc = this.attributes['voc'];
      //		return 	voc && voc[key] != undefined ? voc[key] : key;
      return this.attributes.voc && this.attributes.voc[key];
    },
    switchLang(lang, cb) {
      let 	vocModule = 'voc_en';
      const m = this;

      switch (lang) {
        case 'ru': vocModule = 'voc_ru'; break;
        case 'uk': vocModule = 'voc_uk'; break;
        case 'fr': vocModule = 'voc_fr'; break;
        case 'de': vocModule = 'voc_de'; break;
        case 'jp': vocModule = 'voc_jp'; break;
      }
      require([vocModule], (voc) => {
        m.set({
          lang,
          voc,
        });
        m.save();
        cb && cb();
      });
    },
    helpSegment() { // language segment for help link '/help/%lang%/index.html')
      return this.get('lang') == 'ru' ? 'ru' : 'en';
    },
    save() {
      window.localStorage.settings = JSON.stringify({
        maxCheckFileSize: this.get('maxCheckFileSize'),
        lang: this.get('lang'),
      });
    },
    initialize() {
      this.on('change:maxCheckFileSize', function (m, max) {
        this.save();
      });
      App.xhr.get('/api/10.2/threats', {}, (d, r) => {
        if (Array.isArray(d)) {
          this.set('threats_count', d.length);
        } // else we have troubles
      }, 'json');
    },
  });

  // BASE VIEW
  App.view.base = Backbone.View.extend({
    initialize() {
      this.children = {};
      this.render();
    },
    render(tools) {
      if (typeof (this.template) === 'function') {
        const 	modelJSON = this.model ? this.model.toJSON({ computed: true }) : {};

        modelJSON._tools = tools || this.share;
        this.el.innerHTML = this.template(modelJSON);
      } else if (typeof (this.template) === 'string' && this.template.indexOf('<%') == -1) {
        this.el.innerHTML = this.template;
      } else if (this.template) {
        this.template = _.template(this.template);
        return this.render(tools);
      }

      this.controls = {};
      this._bindByRole();
      return this;
    },
    remove() {
      this.trigger('remove');
      _.each(this.children, (view) => {
        // view.model && view.model.destroy();
        view.remove();
      });
      this.$el.off().empty().remove();
      this.stopListening();
      for (const key in this.controls) {
        delete this.controls[key];
      }
      this.controls = {};
      return this;
    },
    // @memberOf App.view.base
    // @param {String} controlName - destination node
    // @param {Backbone.View} view - child view
    addSubView(controlName, view, setId) {
      if (this.controls[controlName]) {
        this.controls[controlName].append(view.el);
        this.children[setId || view.cid] = view;
        return view;
      }
    },
    // @param {HtmlElement} $target
    _bindByRole($target) {
      const 	CATCH_DEFIS = /-(\w)/g;
      const replaceDefis = function (str, p1) { return p1.toUpperCase(); };
      const roleNodes = ($target || this.el).querySelectorAll('[data-co]');

      // this.controls = {};
      for (var i = 0, len = roleNodes.length, field; i < len; i++) {
        field = roleNodes[i].dataset.co.replace(CATCH_DEFIS, replaceDefis);
        this.controls[field] = $(roleNodes[i]);
      }
    },
  });

  // Base Model for Blocks at main page
  App.model.infoBlock = Backbone.Model.extend({
    methodMap: {
      create: 'POST',
      update: 'PUT',
      delete: 'DELETE',
      read: 'GET',
    },
    sync(method, model, options) {
      // To get original responce: return Backbone.sync(method, model, options);
      const 	type = options.type || this.methodMap[method];
      const url = options.url || this.url;

      return App.xhr[type == 'GET' ? 'get' : 'post'](url, {}, (d, r, ec) => {
        options.success(d ? (options.parse ? options.parse(d) : d) : {}, r);
      }, 'json'); // with forse json parsing!
    },
  });
  /// / * * * Model instance * * *
  App.user = new App.model.auth();
  App.i18n = new App.model.I18n();
  App.licenses = new App.model.lic();
  App.init = function (initSettings) {
    $c.initControlDelegation();
    Backbone.history.start();
  };
  $c._ = function (s) {
    return App.i18n._(s);
  };

  return App;
});
